import six


@six.python_2_unicode_compatible
class JirafsError(Exception):
    def __str__(self):
        super_str = super(JirafsError, self).__str__()
        if not super_str:
            return six.text_type(
                self.__class__.__name__
            )

        return super_str

class NotTicketFolderException(JirafsError):
    pass

class LocalCopyOutOfDate(JirafsError):
    pass

class JiraInteractionFailed(JirafsError):
    pass


class GitCommandError(JirafsError):
    def __init__(self, *args, **kwargs):
        self._command = kwargs.pop('cmd', None)
        self._returncode = kwargs.pop('returncode', None)
        self._output = kwargs.pop('stdout', None)

        super(GitCommandError, self).__init__(*args, **kwargs)

    @property
    def command(self):
        return self._command

    @property
    def returncode(self):
        return self._returncode

    @property
    def output(self):
        return self._output


class IssueValidationError(JirafsError):
    pass


class MacroError(JirafsError):
    pass


class MacroContentError(MacroError):
    pass



from __future__ import print_function

import argparse
import json
import logging
import os
import re
import sys

from blessings import Terminal
import six
from verlib import NormalizedVersion

from . import __version__


logger = logging.getLogger(__name__)


class PluginError(Exception):
    pass


class PluginValidationError(PluginError):
    pass


class PluginOperationError(PluginError):
    pass


class CommandResult(six.text_type):
    def __new__(
        cls, string=None, return_code=None, cursor=0, no_format=False, **kwargs
    ):
        if string is None:
            string = ''
        if string and not string.endswith('\n'):
            string = string + '\n'

        terminal = Terminal()
        if not no_format:
            kwargs['t'] = terminal
            try:
                string = string.format(**kwargs)
            except KeyError:
                logger.warning(
                    "An error was encountered while attempting to format "
                    "string; returning the original string unformatted. "
                    "The caller may want to use the 'no_format' option if "
                    "the outgoing string includes curly braces.",
                )

        self = super(CommandResult, cls).__new__(cls, string)
        self.return_code = return_code
        self.terminal = terminal
        self.cursor = cursor

        return self

    def _echo(self, message):
        print(message, end='')

    def echo(self):
        self._echo(self[self.cursor:])
        self.cursor = len(self)

        return self

    def add_line(self, the_line, no_format=False, **kwargs):
        if not the_line.endswith('\n'):
            the_line = the_line + '\n'

        if not no_format:
            kwargs['t'] = self.terminal
            try:
                the_line = the_line.format(**kwargs)
            except KeyError:
                logger.warning(
                    "An error was encountered while attempting to format "
                    "string; returning the original string unformatted. "
                    "The caller may want to use the 'no_format' option if "
                    "the outgoing string includes curly braces.",
                )

        new_result = CommandResult(the_line)
        return self + new_result

    def __add__(self, other):
        joined_strings = super(CommandResult, self).__add__(other)

        return_code = None
        if self.return_code is not None:
            return_code = self.return_code
        if isinstance(other, CommandResult) and other.return_code is not None:
            return_code = other.return_code

        return CommandResult(
            joined_strings,
            return_code=return_code,
            cursor=self.cursor
        )

    @property
    def return_code(self):
        return self._return_code

    @return_code.setter
    def return_code(self, value):
        self._return_code = int(value) if value is not None else None


class JirafsPluginBase(object):
    MIN_VERSION = None
    MAX_VERSION = None

    def validate(self, **kwargs):
        if not self.MIN_VERSION or not self.MAX_VERSION:
            raise PluginValidationError(
                "Minimum and maximum version numbers not specified."
            )

        min_version = NormalizedVersion(self.MIN_VERSION)
        max_version = NormalizedVersion(self.MAX_VERSION)
        curr_version = NormalizedVersion(__version__)
        if not min_version <= curr_version <= max_version:
            raise PluginValidationError(
                "Plugin '%s' is not compatible with version %s of Jirafs; "
                "minimum version: %s; maximum version %s." % (
                    self.plugin_name,
                    __version__,
                    self.MIN_VERSION,
                    self.MAX_VERSION,
                ),
            )

        return True


class Plugin(JirafsPluginBase):
    def __init__(self, ticketfolder, plugin_name, **kwargs):
        self.ticketfolder = ticketfolder
        self.plugin_name = plugin_name

    @property
    def metadata_filename(self):
        return self.ticketfolder.get_metadata_path(
            'plugin_meta',
            '%s.json' % self.plugin_name,
        )

    def get_configuration(self):
        config = self.ticketfolder.get_config()
        if config.has_section(self.plugin_name):
            return dict(config.items(self.plugin_name))
        return {}

    def get_metadata(self):
        try:
            with open(self.metadata_filename, 'r') as _in:
                return json.loads(_in.read())
        except (IOError, OSError):
            return {}

    def set_metadata(self, data):
        with open(self.metadata_filename, 'w') as out:
            out.write(
                json.dumps(
                    data,
                    indent=4,
                    sort_keys=True,
                )
            )


class CommandPlugin(JirafsPluginBase):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def truncate_field_value(self, original_value, length=30):
        if original_value is None:
            original_value = ""
        elif not isinstance(original_value, six.string_types):
            original_value = six.text_type(original_value)
        value = original_value.strip()
        for newline in ('\n', '\r'):
            if newline in value:
                value = value[0:value.find(newline)]

        value = value[0:length]

        if value != original_value:
            value = value[0:length-1] + u'\u2026'

        return value

    def get_description(self):
        try:
            return self.__doc__.strip()
        except AttributeError:
            raise None

    def add_arguments(self, parser):
        pass

    def parse_arguments(self, parser, extra_args):
        return parser.parse_args(extra_args)

    @classmethod
    def get_command_result(cls, result, original=None):
        if not isinstance(result, CommandResult):
            result = CommandResult(result)

        if original is not None:
            result = original + result

        return result

    @classmethod
    def execute_command(cls, extra_args, jira, path, command_name, **ckwargs):
        from .ticketfolder import TicketFolder
        cmd = cls(
            plugin_name=command_name
        )

        parser = argparse.ArgumentParser(
            prog=os.path.basename(sys.argv[0]) + ' ' + command_name,
            description=cmd.get_description(),
        )
        if cmd.auto_instantiate_folder():
            parser.add_argument(
                '--no-migrate',
                dest='migrate',
                default=True,
                action='store_false'
            )
        cmd.add_arguments(parser)
        args = cmd.parse_arguments(parser, extra_args)

        folder = None
        folder_plugins = []
        if cmd.auto_instantiate_folder():
            folder = TicketFolder(path, jira, migrate=args.migrate)
            folder_plugins = folder.plugins

        kwargs = {
            'args': args,
            'folder': folder,
            'jira': jira,
            'path': path,
            'parser': parser,
        }
        pre_method = 'pre_%s' % command_name
        post_method = 'post_%s' % command_name
        for plugin in folder_plugins:
            if not hasattr(plugin, pre_method):
                continue
            method = getattr(plugin, pre_method)
            result = method(**kwargs)
            if result is not None:
                kwargs = result

        cmd.validate(**kwargs)
        result = cls.get_command_result(
            cmd.handle(**kwargs)
        )

        for plugin in folder_plugins:
            if not hasattr(plugin, post_method):
                continue
            method = getattr(plugin, post_method)
            post_result = method(result)
            if post_result is not None:
                result = cls.get_command_result(
                    post_result,
                    original=result
                )

        if getattr(cls, 'RUN_FOR_SUBTASKS', False):
            for subfolder in folder.subtasks:
                try:
                    cls.execute_command(
                        extra_args,
                        jira,
                        subfolder.path,
                        command_name,
                        **ckwargs
                    )
                except Exception as e:
                    logger.exception(
                        "Exception encountered while running "
                        "'%s' for ticket subfolder '%s': %s" % (
                            command_name,
                            subfolder.ticket_number,
                            e
                        )
                    )

        return result

    def handle(self, *args, **kwargs):
        return self.cmd(*args, **kwargs)

    def cmd(self, *args, **kwargs):
        # By default, no return value; just execute and move along
        self.main(*args, **kwargs)

    def main(self, *args, **kwargs):
        raise NotImplementedError()

    def try_subfolders(self):
        return getattr(self, 'TRY_SUBFOLDERS', False)

    def auto_instantiate_folder(self):
        return getattr(
            self,
            'AUTOMATICALLY_INSTANTIATE_FOLDER',
            True,
        )


class DirectOutputCommandPlugin(CommandPlugin):
    def cmd(self, *args, **kwargs):
        return self.main(*args, **kwargs)


class MacroPlugin(Plugin):
    COMPONENT_NAME = None
    MATCHER = None

    def __init__(self, folder, plugin_name, *args, **kwargs):
        self.ticketfolder = folder
        self.plugin_name = plugin_name
        self._args = args
        self._kwargs = kwargs

    def get_matcher(self):
        return re.compile(
            self.BASE_REGEX.format(tag_name=self.COMPONENT_NAME),
            re.MULTILINE | re.DOTALL
        )

    def get_matches(self, content):
        return self.get_matcher().finditer(content)

    def get_attributes(self, tag):
        if ':' not in tag:
            return {}

        attributes = {}

        attribute_content = tag[tag.find(':')+1:-1]
        for segment in attribute_content.split('|'):
            if '=' not in segment:
                attributes[segment] = True
                continue

            attribute, value = segment.split('=', 1)
            attributes[attribute] = value

        return attributes

    def process_text_data(self, content):
        def run_replacement(match_data):
            data = match_data.groupdict()

            return self.execute_macro(
                data.get('content'),
                **self.get_attributes(data.get('start', ''))
            )

        try:
            return self.get_matcher().sub(run_replacement, content)
        except Exception as e:
            self.ticketfolder.log(
                "Error encountered while running macro %s: %s",
                args=(
                    self.plugin_name,
                    e
                ),
                level=logging.ERROR,
            )

    def execute_macro(self, data, **attributes):
        raise NotImplementedError()


class BlockElementMacroPlugin(MacroPlugin):
    BASE_REGEX = (
        r'^(?P<start>{{{tag_name}[^}}]*}})(?P<content>.*?)'
        r'(?P<end>{{{tag_name}}})$'
    )


class VoidElementMacroPlugin(MacroPlugin):
    BASE_REGEX = (
        r'^(?P<start>{{{tag_name}}})'
    )



import os
import re
import shutil
import subprocess
import tempfile

import six
from six.moves.urllib import parse

from jirafs import constants, exceptions, utils
from jirafs.plugin import CommandPlugin
from jirafs.ticketfolder import TicketFolder


class Command(CommandPlugin):
    """ Clone a new ticketfolder for the specified ticket URL"""
    MIN_VERSION = '1.15'
    MAX_VERSION = '1.99.99'
    AUTOMATICALLY_INSTANTIATE_FOLDER = False

    TICKET_RE = re.compile('.*\/browse\/(\w+-\d+)\/?')

    def handle(self, args, jira, path, **kwargs):
        ticket_url = args.ticket_url[0]

        if not os.path.exists(os.path.realpath(ticket_url)):
            ticket_url_parts = parse.urlparse(ticket_url)
            if not ticket_url_parts.netloc:
                default_server = utils.get_default_jira_server()
                ticket_url = parse.urljoin(
                    default_server,
                    'browse/' + ticket_url + '/'
                )

        path = args.path[0] if args.path else None

        return self.cmd(path, ticket_url, jira)

    def clone_from_issue(self, match, ticket_url, path, jira):
        if not path:
            path = match.group(1)
        path = os.path.realpath(path)
        os.mkdir(path)
        folder = TicketFolder.initialize_ticket_folder(ticket_url, path, jira)

        utils.run_command_method_with_kwargs('pull', folder=folder)

        folder.log(
            "Issue %s cloned successfully to %s",
            (
                folder.issue_url,
                folder.path,
            )
        )

        return folder

    def clone_from_git_repository(self, url, path, jira):
        temp_dir = tempfile.mkdtemp()

        subprocess.check_call(
            [
                'git',
                'clone',
                url,
                temp_dir
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        for branch in ['jira', 'master']:
            subprocess.check_call(
                [
                    'git',
                    'checkout',
                    branch,
                ],
                cwd=temp_dir,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

        issue_url_path = os.path.join(
            temp_dir, constants.METADATA_DIR, 'issue_url'
        )
        with open(issue_url_path, 'r') as issue_url_file:
            issue_url = issue_url_file.read()

        match = self.TICKET_RE.match(issue_url)
        if not match:
            shutil.rmtree(temp_dir)
            raise exceptions.NotTicketFolderException(
                "The git repository at %s is not a Jirafs backup." % (
                    url,
                )
            )

        if path:
            path = os.path.realpath(path)
        else:
            # This will create a subdirectory of the cwd named after
            # the ticket's number
            path = os.path.realpath(match.group(1))

        # Move the temporary clone into the proper place
        os.rename(temp_dir, path)

        # Re-clone the shadow repository
        shadow_path = os.path.join(path, constants.METADATA_DIR, 'shadow')
        subprocess.check_call(
            [
                'git',
                'clone',
                path,
                shadow_path,
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # Move the .git directory to where we hide it
        os.rename(
            os.path.join(path, '.git'),
            os.path.join(path, constants.METADATA_DIR, 'git')
        )

        # Reset the shadow's URL to use a relative path
        subprocess.check_call(
            [
                'git',
                'remote',
                'set-url',
                'origin',
                '../git',
            ],
            cwd=shadow_path,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        subprocess.check_call(
            [
                'git',
                'checkout',
                'jira'
            ],
            cwd=shadow_path,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        folder = TicketFolder(
            path,
            jira,
        )
        folder.run_git_command(
            'config',
            '--file=%s' % folder.get_metadata_path(
                'git', 'config',
            ),
            'core.excludesfile',
            '.jirafs/gitignore',
        )
        folder.log(
            "Cloned Jirafs ticket folder for %s at %s; on hash %s",
            (
                folder.issue_url,
                folder.path,
                folder.run_git_command(
                    'rev-parse',
                    'master',
                )
            )
        )
        return folder

    def main(self, path, url, jira):
        match = self.TICKET_RE.match(url)
        if match:
            return self.clone_from_issue(
                match,
                url,
                path,
                jira,
            )

        # Try checking if it's a git repository, too
        try:
            subprocess.check_call(
                [
                    'git',
                    'ls-remote',
                    url
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            return self.clone_from_git_repository(
                url,
                path,
                jira,
            )
        except subprocess.CalledProcessError:
            pass

        raise exceptions.JirafsError(
            "\'%s\' is neither a valid JIRA ticket URL, "
            "nor Jirafs remote backup" % (
                url
            )
        )

    def add_arguments(self, parser):
        parser.add_argument(
            'ticket_url',
            nargs=1,
            type=six.text_type
        )
        parser.add_argument(
            'path',
            nargs='*',
            type=six.text_type,
        )



import io

from jirafs.plugin import CommandPlugin


class Command(CommandPlugin):
    """ Merge remote changes into your local copy """
    RUN_FOR_SUBTASKS = True
    TRY_SUBFOLDERS = True
    MIN_VERSION = '1.15'
    MAX_VERSION = '1.99.99'

    def handle(self, args, folder, **kwargs):
        return self.cmd(folder, args)

    def add_arguments(self, parser):
        parser.add_argument(
            'action',
            default='get',
            choices=[
                'get',
                'reset',
            ]
        )

    def main(self, folder, args):
        path = folder.get_metadata_path('macros_applied.patch')
        if args.action == 'reset':
            with io.open(path, 'w', encoding='utf-8') as out:
                out.write(u'\n\n')

            print(
                "Macro patch successfully reset. Be sure to run "
                "`jirafs commit` or `jirafs submit` for these changes "
                "to take effect."
            )
        elif args.action == 'get':
            with io.open(path, 'r', encoding='utf-8') as in_:
                print(in_.read())



from textwrap import dedent
from unittest import TestCase

from jirafs.jirafieldmanager import JiraFieldManager


class TestJiraFieldManager(TestCase):
    def test_decode(self):
        encoded_values = dedent("""
            * summary (summary):
                This is a test summary
            * Longer Message (longer_message):
                This is a much
                longer message that happens
                to contain newlines.
        """)

        expected_result = {
            'summary': 'This is a test summary',
            'longer_message': (
                'This is a much\n'
                'longer message that happens\n'
                'to contain newlines.'
            )
        }

        actual_result = JiraFieldManager(
            encoded_values
        )

        self.assertEqual(
            expected_result,
            actual_result
        )

    def test_decode_with_json_values(self):
        encoded_values = dedent("""
            * Dictionary Field (dictionary_field):
                {
                    "name": "Adam Coddington"
                }
            * Integer Field (integer_field):
                10
            * String Field (string_field):
                Hello
            * List Field (list_field):
                [
                    "Alphabet",
                    {
                        "One": "Two"
                    }
                ]
        """)

        expected_result = {
            "dictionary_field": {
                "name": "Adam Coddington"
            },
            "integer_field": 10,
            "string_field": "Hello",
            "list_field": [
                "Alphabet",
                {
                    "One": "Two",
                }
            ]
        }

        actual_result = JiraFieldManager(
            encoded_values
        )

        self.assertEqual(
            expected_result,
            actual_result,
        )



import os
import tempfile

import mock
from mock import patch

from jirafs.utils import run_command_method_with_kwargs

from ..base import BaseTestCase


class BaseCommandTestCase(BaseTestCase):
    def setUp(self):
        super(BaseCommandTestCase, self).setUp()
        self.arbitrary_ticket_number = 'ALPHA-123'
        self.root_folder = tempfile.mkdtemp()
        self.mock_jira = mock.MagicMock()
        self.mock_jira.issue.return_value = self.rehydrate_issue(
            'basic.issue.json'
        )
        self.mock_get_jira = lambda _, config=None: self.mock_jira

        with patch(
            'jirafs.ticketfolder.TicketFolder.get_remotely_changed'
        ) as get_remotely_changed:
            get_remotely_changed.return_value = []
            self.ticketfolder = run_command_method_with_kwargs(
                'clone',
                url='http://arbitrary.com/browse/ALPHA-123',
                jira=self.mock_get_jira,
                path=os.path.join(
                    self.root_folder,
                    self.arbitrary_ticket_number,
                )
            )



from jirafs import constants, utils
from jirafs.plugin import CommandPlugin
from jirafs.jirafieldmanager import JiraFieldManager
from jirafs.jiralinkmanager import JiraLinkManager


class Command(CommandPlugin):
    """ Merge remote changes into your local copy """
    RUN_FOR_SUBTASKS = True
    TRY_SUBFOLDERS = True
    MIN_VERSION = '1.15'
    MAX_VERSION = '1.99.99'

    def main(self, folder, **kwargs):
        with utils.stash_local_changes(folder):
            original_merge_base = folder.git_merge_base
            folder.run_git_command('merge', 'jira')
            final_merge_base = folder.git_merge_base

            new_comments = folder.run_git_command(
                'diff',
                '%s..%s' % (
                    original_merge_base,
                    final_merge_base
                ),
                '--',
                constants.TICKET_COMMENTS,
            ).strip()
            if new_comments:
                folder.log(
                    "New comment(s) have been posted."
                )

            jira_fields = JiraFieldManager.create(
                folder, revision='jira'
            )
            master_fields = JiraFieldManager.create(
                folder, revision=original_merge_base
            )
            for field, values in (jira_fields - master_fields).items():
                folder.log(
                    u"Field {field} changed: \"{fr}\" -> \"{to}\"".format(
                        field=field,
                        fr=self.truncate_field_value(values[0]),
                        to=self.truncate_field_value(values[1])
                    )
                )

            jira_links = JiraLinkManager.create(
                folder, revision='jira'
            )
            master_links = JiraLinkManager.create(
                folder, revision=original_merge_base
            )
            for category in ('issue', 'remote'):
                values_dict = (jira_links - master_links).get(category, {})
                for field, values in values_dict.items():
                    folder.log(
                        u"Link {field} changed: \"{fr}\" -> \"{to}\"".format(
                            field=field,
                            fr=values[0],
                            to=values[1]
                        )
                    )

            if original_merge_base != final_merge_base:
                folder.log(
                    "Merged 'jira' into 'master'; merge-base is now %s" % (
                        final_merge_base
                    )
                )
            return utils.PostStatusResponse(
                original_merge_base == final_merge_base,
                final_merge_base
            )



from __future__ import print_function

import collections
import json
import os
import shutil
import subprocess

from behave import *
from jira.client import JIRA


@given('jirafs is installed and configured')
def installed_and_configured(context):
    pass

@given('a cloned ticket with the following fields')
def cloned_ticket_with_following_fields(context):
    jira_client = JIRA(
        {
            'server': context.integration_testing['url'],
            'verify': False,
            'check_update': False,
        },
        basic_auth=(
            context.integration_testing['username'],
            context.integration_testing['password'],
        )
    )
    
    issue_data = {
        'project': {
            'key': context.integration_testing['project']
        },
        'issuetype': {
            'name': 'Task',
        }
    }
    for row in context.table:
        issue_data[row[0]] = json.loads(row[1])

    issue = jira_client.create_issue(issue_data)

    if not 'cleanup_steps' in context:
        context.cleanup_steps = []
    context.cleanup_steps.append(
        lambda context: issue.delete()
    )
    context.execute_steps(
        u'''
        when the command "jirafs clone {url}" is executed
        and we enter the ticket folder for "{url}"
        '''.format(url=issue.permalink())
    )


@when('the command "{command}" is executed')
def execute_command(context, command):
    command = command.format(**context.integration_testing)

    env = os.environ.copy()
    env['JIRAFS_GLOBAL_CONFIG'] = context.integration_testing['config_path']
    env['JIRAFS_ALLOW_USER_INPUT__BOOL'] = '0'

    proc = subprocess.Popen(
        command.encode('utf-8'),
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env=env,
    )

    stdout, stderr = proc.communicate()

    if not hasattr(context, 'executions'):
        context.executions = collections.deque()

    context.executions.appendleft(
        {
            'command': command,
            'stdout': stdout.decode('utf-8'),
            'stderr': stderr.decode('utf-8'),
            'return_code': proc.returncode,
        }
    )


@when('we enter the ticket folder for "{url}"')
def execute_command(context, url):
    url = url.format(**context.integration_testing)

    os.chdir(os.path.join(os.getcwd(), url.split('/')[-1]))


@then('the directory will contain a file named "{filename}"')
def directory_contains_file(context, filename):
    assert filename in os.listdir('.'), "%s not in folder" % filename


@then('the output will contain the text "{expected}"')
def output_will_contain(context, expected):
    expected = expected.format(**context.integration_testing)

    assert (expected in context.executions[0]['stdout']), "%s not in %s" % (
        expected,
        context.executions[0]['stdout'],
    )


@step('print execution results')
def print_stdout(context):
    print(json.dumps(context.executions[0], indent=4, sort_keys=True))


@step('debugger')
def debugger(context):
    import ipdb
    ipdb.set_trace()



import io

import six

from jirafs import constants, exceptions, utils
from jirafs.plugin import CommandPlugin
from jirafs.utils import run_command_method_with_kwargs


class Command(CommandPlugin):
    """ Push locally-committed changes to JIRA """
    TRY_SUBFOLDERS = True
    MIN_VERSION = '1.15'
    MAX_VERSION = '1.99.99'

    def get_valid_issue_link_types(self, folder):
        if not hasattr(self, '_valid_issue_link_types'):
            data = {}
            for item in folder.jira.issue_link_types():
                data[item.outward.lower()] = ('outward', item)
                data[item.inward.lower()] = ('inward', item)

            self._valid_issue_link_types = data
        return self._valid_issue_link_types

    def validate_issue(self, folder):
        status = folder.status()

        # Validate issue statuses
        valid_types = self.get_valid_issue_link_types(folder)
        links = status['ready']['links'].get('issue', {})
        for target, data in links.items():
            if data[1] is None:
                continue
            if data[1].get('status') not in valid_types:
                raise exceptions.IssueValidationError(
                    u"{status} is not a valid issue link type for {target}; "
                    u"options include the following: {options}".format(
                        status=data[1].get('status'),
                        target=target,
                        options=', '.join(valid_types.keys())
                    )
                )

    def main(self, folder, **kwargs):
        self.validate_issue(folder)
        with utils.stash_local_changes(folder):
            status = folder.status()

            if not folder.is_up_to_date():
                raise exceptions.LocalCopyOutOfDate(
                    "Your local copy is out-of-date.  You must use "
                    "the 'merge' command to update your local copy "
                    "before pushing changes."
                )

            file_meta = folder.get_remote_file_metadata(shadow=False)

            for filename in status['ready']['files']:
                upload = six.BytesIO(
                    folder.get_local_file_at_revision(
                        filename,
                        'HEAD',
                        binary=True
                    )
                )
                filename, upload = folder.execute_plugin_method_series(
                    'alter_file_upload',
                    args=((filename, upload, ), ),
                    single_response=True,
                )
                folder.log(
                    'Uploading file "%s"',
                    (filename, ),
                )
                # Delete the existing issue if there is one
                for attachment in folder.issue.fields.attachment:
                    if attachment.filename == filename:
                        attachment.delete()
                upload.seek(0)
                attachment = folder.jira.add_attachment(
                    folder.ticket_number,
                    upload,
                    filename=filename,
                )
                file_meta[filename] = attachment.created

            folder.set_remote_file_metadata(file_meta, shadow=False)

            comment = folder.get_new_comment(clear=True, ready=True)
            if comment:
                folder.log(
                    u'Adding comment "{comment}"'.format(
                        comment=self.truncate_field_value(comment)
                    )
                )
                folder.jira.add_comment(folder.ticket_number, comment)

            collected_updates = {}
            for field, diff_values in status['ready']['fields'].items():
                collected_updates[field] = diff_values[1]

            if collected_updates:
                folder.log(
                    'Updating fields "%s"',
                    (collected_updates, )
                )
                folder.issue.update(**collected_updates)

            links = status['ready']['links']
            statuses = self.get_valid_issue_link_types(folder)
            for target, data in links.get('issue', {}).items():
                orig = data[0]
                new = data[1]
                other = folder.jira.issue(target)
                if orig is None:
                    # New links
                    status_data = statuses[new['status']]
                    args = [
                        status_data[1].name,
                    ]
                    if status_data[0] == 'inward':
                        args.extend([
                            other,
                            folder.issue
                        ])
                    elif status_data[0] == 'outward':
                        args.extend([
                            folder.issue,
                            other,
                        ])
                    folder.jira.create_issue_link(*args)
                elif new is None:
                    # Deleted links
                    for existing_link in folder.issue.fields.issuelinks:
                        if (
                            hasattr(existing_link, 'inwardIssue') and
                            existing_link.inwardIssue.key == target
                        ):
                            existing_link.delete()
                        if (
                            hasattr(existing_link, 'outwardIssue') and
                            existing_link.outwardIssue.key == target
                        ):
                            existing_link.delete()
                else:
                    # Changed links
                    for existing_link in folder.issue.fields.issuelinks:
                        if (
                            hasattr(existing_link, 'inwardIssue') and
                            existing_link.inwardIssue.key == target
                        ):
                            existing_link.type = statuses[new['status']][1]
                            existing_link.update()
                        if (
                            hasattr(existing_link, 'outwardIssue') and
                            existing_link.outwardIssue.key == target
                        ):
                            existing_link.type = statuses[new['status']][1]
                            existing_link.update()

            links = status['ready']['links']
            remote_links = folder.jira.remote_links(folder.issue)
            # Workaround for bug in python-jira:
            folder.jira._applicationlinks = []
            for target, data in links.get('remote', {}).items():
                orig = data[0]
                new = data[1]
                if orig is None:
                    # New links
                    link_object = {
                        'url': target,
                        'title': new['description'],
                    }
                    folder.jira.add_remote_link(folder.issue, link_object)
                elif new is None:
                    # Deleted links
                    for existing_link in remote_links:
                        if existing_link.object.url == target:
                            existing_link.delete()
                else:
                    # Changed links
                    for existing_link in remote_links:
                        if existing_link.object.url == target:
                            existing_link.update({
                                'url': target,
                                'title': new['description']
                            })

            # Commit local copy
            folder.run_git_command('reset', '--soft', failure_ok=True)
            folder.run_git_command(
                'add',
                folder.get_path('.jirafs/remote_files.json'),
                failure_ok=True
            )
            folder.run_git_command(
                'add',
                folder.get_path(constants.TICKET_NEW_COMMENT),
                failure_ok=True
            )
            folder.run_git_command(
                'commit', '-m', 'Pushed local changes', failure_ok=True
            )

            # Apply macros and record the changes so we can reverse
            # them a little later.
            macro_patch_filename = folder.get_path(
                '.jirafs/macros_applied.patch'
            )
            folder.run_git_command(
                'apply',
                macro_patch_filename,
                failure_ok=True
            )
            fields = folder.get_fields()
            for field, value in collected_updates.items():
                fields[field] = value
            fields.write()

            field_data_files = fields.get_field_data_files()
            result = folder.run_git_command('diff', *field_data_files)
            if result.strip():
                with io.open(macro_patch_filename, 'w', encoding='utf-8') as o:
                    o.write(result)
                    o.write(u'\n\n')
                folder.run_git_command(
                    'add',
                    macro_patch_filename,
                    failure_ok=True
                )
                folder.run_git_command(
                    'commit', '-m', 'Updating macro patch', failure_ok=True
                )
                # Re-set the applied changes we just made above
                folder.run_git_command(
                    'checkout', '--', '.'
                )
            else:
                with io.open(macro_patch_filename, 'w', encoding='utf-8') as o:
                    o.write(u'\n\n')

            # Commit changes to remote copy, too, so we record remote
            # file metadata.
            folder.run_git_command('fetch', shadow=True)
            folder.run_git_command('merge', 'origin/master', shadow=True)
            folder.run_git_command('add', '-A', shadow=True)
            folder.run_git_command(
                'commit', '-m', 'Pulled remote changes',
                failure_ok=True, shadow=True
            )
            folder.run_git_command('push', 'origin', 'jira', shadow=True)
            pull_result = run_command_method_with_kwargs('pull', folder=folder)
            return pull_result[1]



from jirafs import utils
from jirafs.plugin import CommandPlugin


class Command(CommandPlugin):
    """Create a subtask of a given issue."""
    MIN_VERSION = '1.15'
    MAX_VERSION = '1.99.99'

    def main(self, folder, args, **kwargs):
        summary = ' '.join(args.summary)

        folder.jira.create_issue(
            fields={
                'project': {
                    'key': folder.issue.fields.project.key,
                },
                'summary': summary,
                'issuetype': {
                    'name': 'Sub-task',
                },
                'parent': {
                    'id': folder.issue.key,
                }
            }
        )

        commands = utils.get_installed_commands()
        jira = utils.lazy_get_jira()
        commands['fetch'].execute_command(
            [],
            jira=jira,
            path=folder.path,
            command_name='fetch',
        )

    def add_arguments(self, parser):
        parser.add_argument(
            "summary",
            nargs='+'
        )



from jirafs.plugin import CommandPlugin, CommandResult


class Command(CommandPlugin):
    """ Commit local changes for later submission to JIRA """
    MIN_VERSION = '1.16'
    MAX_VERSION = '1.99.99'

    def main(self, args, folder, **kwargs):
        return folder.process_plugin_builds()

    def cmd(self, *args, **kwargs):
        data = self.main(*args, **kwargs)

        result = CommandResult()
        for key, value in data.items():
            result = result.add_line(
                'Ran build plugin {plugin}',
                plugin=key,
            )
            if not value:
                continue

            if not isinstance(value, basestring):
                value = json.dumps(
                    value,
                    indent=4,
                    sort_keys=True,
                )

            for line in value.split('\n'):
                if not line.strip():
                    continue

                result = result.add_line(
                    '\t{line}',
                    line=line,
                )

        return result




import codecs
import datetime
import fnmatch
import io
import json
import logging
import os
import re
import subprocess

from jira.resources import Issue
import six
from six.moves.urllib import parse

from . import constants
from . import exceptions
from . import migrations
from . import utils
from .jiralinkmanager import JiraLinkManager
from .jirafieldmanager import JiraFieldManager
from .plugin import MacroPlugin, PluginValidationError


logger = logging.getLogger(__name__)


class TicketFolder(object):
    def __init__(self, path, jira, migrate=True, quiet=False):
        self.path = os.path.realpath(
            os.path.expanduser(path)
        )
        self.quiet = quiet
        self.issue_url = self.get_ticket_url()
        self.get_jira = jira

        if not os.path.isdir(self.metadata_dir):
            raise exceptions.NotTicketFolderException(
                "%s is not a synchronizable ticket folder" % (
                    path
                )
            )

        self.plugins = self.load_plugins()

        if migrate:
            self.run_migrations()

        # If no `new_comment.jira.txt` file exists, let's create one
        comment_path = self.get_local_path(constants.TICKET_NEW_COMMENT)
        if not os.path.exists(comment_path):
            with io.open(comment_path, 'w', encoding='utf-8') as out:
                out.write(six.text_type(''))

        # Let's update the ignore file while we're here.
        self.build_ignore_files()

    def __repr__(self):
        if six.PY3:
            value = self.__unicode__()
        else:
            value = self.__str__()
        return "<%s>" % value

    def __unicode__(self):
        return "[%s] at %s" % (self.ticket_number, self.path)

    def __str__(self):
        if six.PY3:
            return self.__unicode__()
        return self.__unicode__().encode('utf8', 'replace')

    @property
    def subtasks(self):
        if hasattr(self, '_subtasks'):
            return self._subtasks
        self._subtasks = []

        subtasks_path = self.get_metadata_path('subtasks')
        if not os.path.exists(subtasks_path):
            return self._subtasks

        with open(subtasks_path, 'r') as in_:
            for line in in_:
                ticket_number = line.strip()
                folder = self.__class__(
                    self.get_path(
                        ticket_number,
                    ),
                    utils.lazy_get_jira()
                )
                self._subtasks.append(folder)

        return self._subtasks

    def execute_plugin_method_series(
        self, name, args=None, kwargs=None, single_response=False
    ):
        if args is None:
            args = []
            use_kwargs = True
        if kwargs is None:
            kwargs = {}
            use_kwargs = False

        if use_kwargs and single_response:
            raise RuntimeError(
                "When executing plugins in series using `single` response "
                "mode, you must specify only args."
            )
        elif args and kwargs:
            raise RuntimeError(
                "Plugins can be ran in series using either args or "
                "kwargs, not both."
            )

        for plugin in self.plugins:
            if not hasattr(plugin, name):
                continue
            method = getattr(plugin, name)
            plugin_result = method(*args, **kwargs)
            if plugin_result is not None:
                if use_kwargs:
                    kwargs = plugin_result
                elif single_response:
                    args = (plugin_result, )
                else:
                    args = plugin_result

        if use_kwargs:
            return kwargs
        elif single_response:
            return args[0]
        return args

    def load_plugins(self):
        config = self.get_config()
        plugins = []

        if not config.has_section(constants.CONFIG_PLUGINS):
            return plugins

        installed_plugins = utils.get_installed_plugins()

        for name, status in config.items(constants.CONFIG_PLUGINS):
            if not utils.convert_to_boolean(status):
                # This plugin is not turned on.
                continue
            if name not in installed_plugins:
                # This plugin is not installed.
                self.log(
                    "Plugin '%s' is not available.",
                    (name, ),
                )
                continue

            plugin = installed_plugins[name](self, name)

            try:
                plugin.validate()
            except PluginValidationError as e:
                self.log(
                    "Plugin '%s' did not pass validation; not loading: %s.",
                    (name, e,)
                )

            plugins.append(plugin)

        return plugins

    def get_config(self):
        local_config_file = self.get_metadata_path('config')
        additional_configs = []
        if os.path.exists(local_config_file):
            additional_configs.append(
                local_config_file
            )

        return utils.get_config(additional_configs)

    def set_config_value(self, section, key, value):
        with utils.stash_local_changes(self):
            local_config_file = self.get_metadata_path('config')

            config = utils.get_config(
                additional_configs=[
                    local_config_file,
                ],
                include_global=False,
            )
            if not config.has_section(section):
                config.add_section(section)
            config.set(section, key, value)

            with open(local_config_file, 'w') as out:
                config.write(out)

            self.run_git_command('add', '.jirafs/config')
            self.run_git_command(
                'commit', '-m', 'Config change',
                failure_ok=True
            )

    @property
    def jira_base(self):
        match = re.match('(.*)\/browse\/.*', self.issue_url)
        if not match:
            raise ValueError(
                "Could not infer JIRA server URL from issue URL %s" % (
                    self.issue_url,
                )
            )
        return match.group(1)

    @property
    def ticket_number(self):
        parts = parse.urlparse(self.issue_url)
        match = re.match('.*\/browse\/(\w+-\d+)\/?.*', parts.path)
        if not match:
            raise ValueError(
                "Could not infer ticket number from URL %s" % self.issue_url
            )
        return match.group(1)

    @property
    def jira(self):
        if not hasattr(self, '_jira'):
            self._jira = self.get_jira(
                self.jira_base,
                config=self.get_config()
            )
        return self._jira

    @property
    def issue(self):
        if not hasattr(self, '_issue'):
            self._issue = self.jira.issue(self.ticket_number)
        return self._issue

    def clear_cache(self):
        if hasattr(self, '_issue'):
            del self._issue
        if hasattr(self, '_jira'):
            del self._jira

    def store_cached_issue(self, shadow=True):
        storable = {
            'options': self.issue._options,
            'raw': self.issue.raw
        }
        with io.open(
            self.get_path('.jirafs/issue.json', shadow=shadow),
            'w',
            encoding='utf-8',
        ) as out:
            out.write(
                six.text_type(
                    json.dumps(
                        storable,
                        indent=4,
                        sort_keys=True,
                        ensure_ascii=False,
                    )
                )
            )

    @property
    def cached_issue(self):
        if not hasattr(self, '_cached_issue'):
            try:
                issue_path = self.get_metadata_path('issue.json')
                with io.open(issue_path, 'r', encoding='utf-8') as _in:
                    storable = json.loads(_in.read())
                    self._cached_issue = Issue(
                        storable['options'],
                        None,
                        storable['raw'],
                    )
            except IOError:
                self.log(
                    'Error encountered while loading cached issue!',
                    level=logging.ERROR,
                )
                self._cached_issue = self.issue
        return self._cached_issue

    @property
    def metadata_dir(self):
        return os.path.join(
            self.path,
            constants.METADATA_DIR,
        )

    @property
    def git_merge_base(self):
        return self.run_git_command(
            'merge-base', 'master', 'jira',
        )

    def get_ticket_url(self):
        try:
            with io.open(
                self.get_metadata_path('issue_url'),
                'r',
                encoding='utf-8'
            ) as in_:
                return in_.read().strip()
        except (IOError, OSError):
            return None

    def get_metadata_path(self, *args):
        return os.path.join(
            self.metadata_dir,
            *args
        )

    def get_remote_file_metadata(self, shadow=True):
        remote_files = self.get_path(
            '.jirafs/remote_files.json',
            shadow=shadow
        )
        try:
            with io.open(remote_files, 'r', encoding='utf-8') as _in:
                data = json.loads(_in.read())
        except IOError:
            data = {}

        return self.execute_plugin_method_series(
            'alter_get_remote_file_metadata',
            args=(data, ),
            single_response=True,
        )

    def set_remote_file_metadata(self, data, shadow=True):
        data = self.execute_plugin_method_series(
            'alter_set_remote_file_metadata',
            args=(data, ),
            single_response=True,
        )
        remote_files = self.get_path(
            '.jirafs/remote_files.json',
            shadow=shadow
        )
        with io.open(remote_files, 'w', encoding='utf-8') as out:
            out.write(
                six.text_type(
                    json.dumps(
                        data,
                        indent=4,
                        sort_keys=True,
                        ensure_ascii=False,
                    )
                )
            )

    def get_local_path(self, *args):
        return os.path.join(
            self.path,
            *args
        )

    def get_shadow_path(self, *args):
        return os.path.join(
            self.get_metadata_path('shadow'),
            *args
        )

    def get_path(self, *args, **kwargs):
        shadow = kwargs.get('shadow', False)

        if shadow:
            return self.get_shadow_path(*args)
        return self.get_local_path(*args)

    @property
    def version(self):
        try:
            with io.open(
                self.get_metadata_path('version'),
                'r',
                encoding='utf-8'
            ) as _in:
                return int(_in.read().strip())
        except IOError:
            return 1

    @property
    def log_path(self):
        return self.get_metadata_path(constants.TICKET_OPERATION_LOG)

    @classmethod
    def initialize_ticket_folder(cls, ticket_url, path, jira):
        path = os.path.realpath(path)

        metadata_path = os.path.join(
            path,
            constants.METADATA_DIR,
        )
        os.mkdir(metadata_path)

        with io.open(
            os.path.join(metadata_path, 'issue_url'),
            'w',
            encoding='utf-8'
        ) as out:
            out.write(six.text_type(ticket_url))

        # Create bare git repository so we can easily detect changes.
        subprocess.check_call(
            (
                'git',
                '--bare',
                'init',
                os.path.join(
                    metadata_path,
                    'git',
                )
            ),
            stdout=subprocess.PIPE
        )
        subprocess.check_call((
            'git',
            'config',
            '--file=%s' % os.path.join(
                metadata_path,
                'git',
                'config'
            ),
            'core.excludesfile',
            constants.GIT_IGNORE_FILE,
        ))

        excludes_path = os.path.join(metadata_path, 'git', 'info', 'exclude')
        with io.open(excludes_path, 'w', encoding='utf-8') as gitignore:
            gitignore.write(
                six.text_type('\n').join(
                    [
                        '%s/git' % constants.METADATA_DIR,
                        '%s/shadow' % constants.METADATA_DIR,
                        '%s/operation.log' % constants.METADATA_DIR,
                    ]
                )
            )

        instance = cls(path, jira, migrate=False)
        instance.log(
            'Ticket folder for issue %s created at %s',
            (instance.ticket_number, instance.path, )
        )
        instance.run_git_command(
            'add', '-A'
        )
        instance.run_git_command(
            'commit', '--allow-empty', '-m', 'Initialized'
        )
        instance.run_migrations(init=True)

        comment_path = instance.get_local_path(constants.TICKET_NEW_COMMENT)
        with io.open(comment_path, 'w', encoding='utf-8') as out:
            out.write(six.text_type(''))

        return instance

    def applied_macros_exist(self, shadow=False):
        macro_patch_filename = self.get_path(
            '.jirafs/macros_applied.patch',
            shadow=shadow,
        )

        if not os.path.exists(macro_patch_filename):
            return False

        with open(macro_patch_filename) as incoming:
            if incoming.read().strip():
                return True

        return False

    def run_git_command(self, command, *args, **kwargs):
        failure_ok = kwargs.get('failure_ok', False)
        shadow = kwargs.get('shadow', False)
        binary = kwargs.get('binary', False)
        stdin = kwargs.get('stdin', '')

        args = list(args)

        if not shadow:
            work_tree = self.path,
            git_dir = self.get_metadata_path('git')
            cwd = self.path
        else:
            work_tree = self.get_metadata_path('shadow')
            git_dir = self.get_metadata_path('shadow/.git')
            cwd = self.get_metadata_path('shadow')

        cmd = [
            'git',
            '--work-tree=%s' % work_tree,
            '--git-dir=%s' % git_dir,
        ]
        cmd.append(command)
        if command == 'commit':
            args.append(
                "--author='%s'" % constants.GIT_AUTHOR
            )
        cmd.extend(args)

        self.log(
            'Executing git command `%s`',
            (
                " ".join(cmd),
            ),
            logging.DEBUG
        )

        handle = subprocess.Popen(
            cmd,
            cwd=cwd,
            stdin=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE
        )
        result, _ = handle.communicate(stdin)

        if handle.returncode != 0 and not failure_ok:
            command = ' '.join(cmd)
            raise exceptions.GitCommandError(
                "Error running command `%s`" % command,
                returncode=handle.returncode,
                stdout=result,
                cmd=command
            )
        if not binary:
            return result.decode('utf-8').strip()
        return result

    def get_local_file_at_revision(
        self, path, revision, failure_ok=True, binary=False
    ):
        return self.run_git_command(
            'show', '%s:%s' % (
                revision,
                path,
            ),
            failure_ok=failure_ok,
            binary=binary,
        )

    def get_ignore_globs(self, which=constants.LOCAL_ONLY_FILE):
        all_globs = [
            constants.TICKET_DETAILS,
            constants.TICKET_COMMENTS,
            constants.TICKET_NEW_COMMENT,
            constants.TICKET_LINKS,
        ]
        for field in constants.FILE_FIELDS:
            all_globs.append(
                constants.TICKET_FILE_FIELD_TEMPLATE.format(field_name=field)
            )

        for plugin in self.plugins:
            if hasattr(plugin, 'get_ignore_globs'):
                all_globs.extend(plugin.get_ignore_globs())

        def get_globs_from_file(input_file):
            globs = []
            for line in input_file.readlines():
                if line.startswith('#') or not line.strip():
                    continue
                globs.append(line.strip())
            return globs

        try:
            with io.open(
                self.get_local_path(which),
                'r',
                encoding='utf-8'
            ) as local_ign:
                all_globs.extend(
                    get_globs_from_file(local_ign)
                )
        except IOError:
            pass

        try:
            with io.open(
                os.path.expanduser('~/%s' % which),
                'r',
                encoding='utf-8'
            ) as global_ignores:
                all_globs.extend(
                    get_globs_from_file(global_ignores)
                )
        except IOError:
            pass

        return all_globs

    def file_matches_globs(self, filename, ignore_globs):
        for glob in ignore_globs:
            if fnmatch.fnmatch(filename, glob):
                return True
        return False

    def get_ready_changes(self):
        ready = {
            'fields': (
                self.get_fields('HEAD') - self.get_fields(self.git_merge_base)
            ),
            'links': (
                self.get_links('HEAD') - self.get_links(self.git_merge_base)
            ),
            'new_comment': self.get_new_comment(ready=True),
        }

        changed_files = self.filter_ignored_files(
            self.run_git_command(
                'diff',
                '--name-only',
                '%s..master' % self.git_merge_base,
            ).split('\n'),
            constants.LOCAL_ONLY_FILE
        )
        ready['files'] = changed_files

        return ready

    def get_uncommitted_changes(self):
        uncommitted = {
            'fields': self.get_fields() - self.get_fields('HEAD'),
            'new_comment': self.get_new_comment(ready=False),
            'links': self.get_links() - self.get_links('HEAD'),
        }

        new_files = self.run_git_command(
            'ls-files', '-o', failure_ok=True
        ).split('\n')
        modified_files = self.run_git_command(
            'ls-files', '-m', failure_ok=True
        ).split('\n')
        uncommitted['files'] = self.filter_ignored_files(
            [
                filename for filename in new_files + modified_files
                if filename
            ],
            constants.LOCAL_ONLY_FILE,
            constants.GIT_IGNORE_FILE,
            constants.GIT_EXCLUDE_FILE,
        )

        return uncommitted

    def get_local_uncommitted_changes(self):
        new_files = self.run_git_command(
            'ls-files', '-o', failure_ok=True
        ).split('\n')
        modified_files = self.run_git_command(
            'ls-files', '-m', failure_ok=True
        ).split('\n')

        committable = self.filter_ignored_files(
            [
                filename for filename in new_files + modified_files if filename
            ],
            constants.LOCAL_ONLY_FILE
        )
        uncommitted = self.filter_ignored_files(
            [
                filename for filename in modified_files + new_files
                if filename not in committable
            ],
            constants.GIT_IGNORE_FILE,
            constants.GIT_EXCLUDE_FILE,
        )

        return {
            'files': uncommitted
        }

    def get_remotely_changed(self):
        metadata = self.get_remote_file_metadata(shadow=True)

        assets = []
        attachments = self.filter_ignored_files(
            getattr(self.issue.fields, 'attachment', []),
            constants.REMOTE_IGNORE_FILE
        )
        for attachment in attachments:
            changed = metadata.get(attachment.filename) != attachment.created
            if changed:
                assets.append(attachment.filename)

        return self.execute_plugin_method_series(
            name='alter_remotely_changed',
            args=(assets, ),
            single_response=True,
        )

    def filter_ignored_files(self, files, *which):
        if len(which) < 1:
            which = [constants.LOCAL_ONLY_FILE]
        if not isinstance(which, (list, tuple)):
            which = [which]

        for list_path in which:
            ignore_globs = self.get_ignore_globs(list_path)

            assets = []
            for fileish in files:
                # Get the actual filename; this is a little gross -- apologies.
                filename = fileish
                attachment = False
                if not isinstance(fileish, six.string_types):
                    filename = fileish.filename
                    attachment = True

                if self.file_matches_globs(filename, ignore_globs):
                    continue
                if (
                    not attachment and
                    not os.path.isfile(os.path.join(self.path, filename))
                ):
                    continue
                if filename.startswith('.'):
                    continue
                assets.append(fileish)
            files = assets

        return self.execute_plugin_method_series(
            name='alter_filter_ignored_files',
            args=(assets, ),
            single_response=True,
        )

    def get_macro_plugins(self):
        if not hasattr(self, '_macro_plugins'):
            config = self.get_config()
            plugins = []

            if not config.has_section(constants.CONFIG_PLUGINS):
                return plugins

            installed_plugins = utils.get_installed_plugins(MacroPlugin)

            for name, status in config.items(constants.CONFIG_PLUGINS):
                if not utils.convert_to_boolean(status):
                    # This plugin is not turned on.
                    continue
                if name not in installed_plugins:
                    # This plugin is not installed.
                    self.log(
                        "Macro plugin '%s' is not available; "
                        "this is probably because this plugin is not a "
                        "macro.",
                        (name, ),
                        level=logging.DEBUG
                    )
                    continue

                plugin = installed_plugins[name](self, name)

                try:
                    plugin.validate()
                except PluginValidationError as e:
                    self.log(
                        "Plugin '%s' did not pass validation; "
                        "not loading: %s.",
                        (name, e,),
                    )

                plugins.append(plugin)

            self._macro_plugins = plugins

        return self._macro_plugins

    def process_plugin_builds(self):
        results = {}

        for plugin in self.plugins:
            if hasattr(plugin, 'run_build_process'):
                results[plugin.plugin_name] = plugin.run_build_process()

        return results

    def process_macros(self, data):
        macro_plugins = self.get_macro_plugins()

        for cls in macro_plugins:
            if isinstance(data, six.string_types):
                data = cls.process_text_data(data)
            else:
                continue

        return data

    def get_links(self, revision=None, path=None):
        kwargs = {}
        if not revision:
            kwargs['path'] = path if path else self.path
        else:
            kwargs['revision'] = revision
        return JiraLinkManager.create(
            self,
            **kwargs
        )

    def get_fields(self, revision=None, path=None):
        kwargs = {}
        if not revision:
            kwargs['path'] = path if path else self.path
        else:
            kwargs['revision'] = revision
        return JiraFieldManager.create(
            self,
            **kwargs
        )

    def get_new_comment(self, clear=False, staged=False, ready=True):
        try:
            with io.open(
                self.get_local_path(constants.TICKET_NEW_COMMENT),
                'r+',
                encoding='utf-8',
            ) as c:
                local_contents = c.read().strip()
            if ready:
                contents = self.get_local_file_at_revision(
                    constants.TICKET_NEW_COMMENT,
                    'HEAD'
                )
                if contents:
                    contents = contents.strip()
                else:
                    contents = ''
            else:
                contents = local_contents

            if not ready and contents == self.get_new_comment(ready=True):
                contents = ''

            if contents == local_contents and clear:
                with io.open(
                    self.get_local_path(constants.TICKET_NEW_COMMENT),
                    'r+',
                    encoding='utf-8',
                ) as c:
                    c.truncate()
        except IOError:
            contents = ''

        # Apply macro plugins
        contents = self.process_macros(contents)

        return self.execute_plugin_method_series(
            name='alter_new_comment',
            args=(contents, ),
            single_response=True,
        )

    def is_up_to_date(self):
        jira_commit = self.run_git_command('rev-parse', 'jira')
        master_commit = self.run_git_command('rev-parse', 'master')

        try:
            self.run_git_command(
                'merge-base', '--is-ancestor', jira_commit, master_commit,
            )
        except exceptions.GitCommandError:
            return False
        return True

    def status(self):
        status = {
            'ready': self.get_ready_changes(),
            'local_uncommitted': self.get_local_uncommitted_changes(),
            'uncommitted': self.get_uncommitted_changes(),
            'up_to_date': self.is_up_to_date(),
        }

        return self.execute_plugin_method_series(
            'alter_status_dict',
            args=(status, ),
            single_response=True,
        )

        return status

    def run_migrations(self, init=False):
        loglevel = logging.INFO
        if init:
            loglevel = logging.DEBUG
        else:
            if not os.path.exists(self.get_metadata_path('git')):
                raise exceptions.JirafsError(
                    u"{path} is not a valid ticket folder!".format(
                        path=self.path
                    )
                )

            if self.version < constants.CURRENT_REPO_VERSION:
                print(
                    u"Your ticket folder at {path} is out-of-date "
                    u"and is being automatically updated.".format(
                        path=self.path
                    )
                )
        while self.version < constants.CURRENT_REPO_VERSION:
            migrator = getattr(
                migrations,
                'migration_%s' % str(self.version + 1).zfill(4)
            )
            self.migrate(migrator, loglevel=loglevel, init=init)

    def migrate(self, migrator, loglevel=logging.INFO, init=False):
        with utils.stash_local_changes(self):
            self.log('%s: Migration started', (migrator.__name__, ), loglevel)
            migrator(self, init=init)
            self.log('%s: Migration finished', (migrator.__name__, ), loglevel)

    def build_ignore_files(self):
        metadata_excludes = [
            'git',
            'shadow',
            'operation.log',
            'subtasks',
        ]
        with codecs.open(
            self.get_local_path(constants.GIT_EXCLUDE_FILE),
            'w',
            'utf-8'
        ) as out:
            for line in metadata_excludes:
                out.write(
                    '%s/%s\n' % (
                        constants.METADATA_DIR,
                        line,
                    )
                )
            subtask_list_path = self.get_metadata_path('subtasks')
            if os.path.exists(subtask_list_path):
                with open(subtask_list_path, 'r') as in_:
                    for line in in_:
                        out.write(
                            '%s/*\n' % line.strip()
                        )

        with codecs.open(
            self.get_metadata_path('combined_ignore'),
            'w',
            'utf-8'
        ) as out:
            try:
                out.write('# ~/%s\n' % constants.GIT_IGNORE_FILE_PARTIAL)
                with codecs.open(
                    os.path.expanduser(
                        '~/%s' % constants.GIT_IGNORE_FILE_PARTIAL
                    ),
                    'r',
                    'utf-8'
                ) as in_:
                    for line in in_:
                        out.write(
                            '%s\n' % line.strip()
                        )
            except:
                pass

            try:
                out.write(
                    '# %s\n' % (
                        self.get_path(constants.GIT_IGNORE_FILE_PARTIAL)
                    )
                )
                with codecs.open(
                    self.get_path(constants.GIT_IGNORE_FILE_PARTIAL),
                    'r',
                    'utf-8'
                ) as in_:
                    for line in in_:
                        out.write('%s\n' % line.strip())
            except:
                pass

    def log(self, message, args=None, level=logging.INFO):
        if args is None:
            args = []
        logger.log(level, message, *args)
        with io.open(self.log_path, 'a', encoding='utf-8') as log_file:
            log_file.write(
                six.text_type(u"%s\t%s\t%s\n") % (
                    datetime.datetime.utcnow().isoformat(),
                    logging.getLevelName(level),
                    (message % args).replace('\n', '\\n')
                )
            )
        if level >= logging.INFO and not self.quiet:
            print(
                u"[%s %s] %s" % (
                    logging.getLevelName(level),
                    self.issue,
                    message % args
                )
            )

    def get_log(self):
        with io.open(self.log_path, 'r', encoding='utf-8') as log_file:
            return log_file.read()



import os
import shutil
import tempfile
import textwrap


def before_scenario(context, *args, **kwargs):
    context.starting_dir = os.getcwd()
    context.temp_dir = tempfile.mkdtemp()
    os.chdir(context.temp_dir)

    keys = {
        'known_ticket_url': 'INTEGRATION_TESTING_KNOWN_TICKET',
        'username': 'INTEGRATION_TESTING_USERNAME',
        'url': 'INTEGRATION_TESTING_URL',
        'project': 'INTEGRATION_TESTING_PROJECT',
        'password': 'INTEGRATION_TESTING_PASSWORD',
    }
    context.integration_testing = {}
    for k, v in keys.items():
        context.integration_testing[k] = os.environ[v]

    context.integration_testing['config_path'] = (
        os.path.join(os.getcwd(), 'jirafs_config')
    )

    with open(context.integration_testing['config_path'], 'w') as out:
        out.write(
            textwrap.dedent(
                """\
                    [{url}]
                    username={username}
                    password={password}

                    [jira]
                    server={url}

                """.format(
                    **context.integration_testing
                )
            )
        )


def after_scenario(context, *args, **kwargs):
    os.chdir(context.starting_dir)
    shutil.rmtree(context.temp_dir)

    if hasattr(context, 'cleanup_steps'):
        for step in context.cleanup_steps:
            step(context)



import six

from jirafs.plugin import CommandPlugin


class Command(CommandPlugin):
    """Assign the current task to a user"""
    MIN_VERSION = '1.15'
    MAX_VERSION = '1.99.99'

    def main(self, args, folder, **kwargs):
        username = args.username
        if not username:
            username = folder.get_config().get(
                folder.jira_base,
                'username',
            )
        folder.jira.assign_issue(folder.issue, username)
        folder.log(
            'Successfully assigned %s to %s.',
            args=(
                folder.issue.key,
                username
            ),
        )

    def add_arguments(self, parser):
        parser.add_argument(
            'username',
            nargs='?',
            type=six.text_type
        )



import json
import os
import subprocess

from jirafs.exceptions import JirafsError
from jirafs.plugin import CommandPlugin, CommandResult
from jirafs.utils import run_command_method_with_kwargs


class Command(CommandPlugin):
    """ Check whether a given dotpath matches an expected value """
    TRY_SUBFOLDERS = True
    MIN_VERSION = '1.15'
    MAX_VERSION = '1.99.99'

    def handle(self, args, folder, **kwargs):
        return self.cmd(
            folder,
            args.field_name,
            args.field_value,
            isjson=args.json,
            negate=args.negate,
            raw=args.raw,
            quiet=args.quiet,
            execute=args.execute,
            execute_here=args.execute_here,
        )

    def add_arguments(self, parser):
        parser.add_argument(
            'field_name',
        )
        parser.add_argument(
            'field_value',
        )
        parser.add_argument(
            '--json',
            help=(
                'Process the provided field value as JSON'
            ),
            action='store_true',
            default=False
        )
        parser.add_argument(
            '--execute',
            help=(
                'Execute a command for each matching result; by default,'
                ' will be executed from within the matching folder directory'
                ' when executed on a folder containing multiple ticket'
                ' folders.  See --execute-here to change this behavior.'
                ' The string {} will be replaced with the folder directory '
                ' path.'
            ),
            default=None,
        )
        parser.add_argument(
            '--execute-here',
            help=(
                'Do not switch directories to matching folders\'s paths'
                ' when using --execute.'
            ),
            action='store_true',
            default=False
        )
        parser.add_argument(
            '--negate',
            help=(
                'Compare the field value without applying '
                'plugin transformations'
            ),
            action='store_true',
            default=False
        )
        parser.add_argument(
            '--raw',
            help=(
                'Return the field value without applying '
                'plugin transformations'
            ),
            action='store_true',
            default=False
        )
        parser.add_argument(
            '--quiet',
            help=(
                'Print no message to stdout indicating success or failure'
            ),
            action='store_true',
            default=False
        )

    def main(
        self, folder, field_name, field_value, isjson,
        negate, raw, quiet, execute, execute_here
    ):
        actual_value = run_command_method_with_kwargs(
            'field',
            method='get_field_value_by_dotpath',
            folder=folder,
            field_name=field_name,
            raw=raw,
        )

        if isjson:
            field_value = json.loads(field_value)

        success = actual_value == field_value

        comparison_result = u" != "
        if success:
            comparison_result = u" == "
        message = u"{left} {comparison} {right}".format(
            left=actual_value,
            comparison=comparison_result,
            right=field_value,
        )

        if negate:
            success = not success

        if execute and success:
            execute = execute.replace('{}', folder.path)
            subprocess.call(
                execute,
                shell=True,
                cwd=os.getcwd() if execute_here else folder.path
            )

        return (
            message if not quiet else None,
            0 if success else 1,
        )

    def cmd(self, *args, **kwargs):
        message, return_code = self.main(*args, **kwargs)

        return CommandResult(message, return_code=return_code)



import json

from jirafs.exceptions import JirafsError
from jirafs.plugin import CommandPlugin


class Command(CommandPlugin):
    """ Get the status of the current ticketfolder """
    TRY_SUBFOLDERS = True
    MIN_VERSION = '1.15'
    MAX_VERSION = '1.99.99'

    def handle(self, args, folder, **kwargs):
        return self.cmd(
            folder, args.field_name, args.value, as_json=args.json
        )

    def add_arguments(self, parser):
        parser.add_argument(
            'field_name',
        )
        parser.add_argument(
            'value',
        )
        parser.add_argument(
            '--json',
            help=(
                'Interpret value argument as JSON data.'
            ),
            action='store_true',
            default=False
        )

    def main(self, folder, field_name, value, as_json=False):
        fields = folder.get_fields()

        if as_json:
            try:
                value = json.loads(value)
            except ValueError:
                raise JirafsError(
                    "Value '%s' could not be decoded as JSON." % (
                        value,
                    )
                )

        key_dotpath = None
        if '.' in field_name:
            field_name, key_dotpath = field_name.split('.', 1)

        if field_name not in fields:
            raise JirafsError("Field '%s' does not exist." % field_name)

        if key_dotpath:
            data = fields[field_name]
            try:
                cursor = data
                dotpath_parts = key_dotpath.split('.')
                last_key = len(dotpath_parts) - 1
                for idx, component in enumerate(dotpath_parts):
                    if idx == last_key:
                        cursor[component] = value
                        break
                    elif not isinstance(cursor.get(component), dict):
                        raise JirafsError(
                            "Key '%s' (of dotpath '%s') is not an object "
                            "in field '%s'." % (
                                component,
                                key_dotpath,
                                field_name,
                            )
                        )
                    else:
                        if component not in data:
                            raise JirafsError(
                                "Key '%s' (of dotpath '%s') could not be found "
                                "in field '%s'." % (
                                    component,
                                    key_dotpath,
                                    field_name,
                                )
                            )
                        cursor = cursor[component]
            except (ValueError, TypeError):
                raise JirafsError(
                    "Field '%s' could not be parsed as JSON for retrieving "
                    "dotpath '%s'." % (
                        field_name,
                        key_dotpath,
                    )
                )
            value = data
        else:
            data = value

        fields[field_name] = data
        fields.write()



import json

from jirafs.exceptions import JirafsError
from jirafs.plugin import CommandPlugin, CommandResult


class Command(CommandPlugin):
    """ Get the status of the current ticketfolder """
    TRY_SUBFOLDERS = True
    MIN_VERSION = '1.15'
    MAX_VERSION = '1.99.99'

    def handle(self, args, folder, **kwargs):
        return self.cmd(
            folder, args.field_name, raw=args.raw, formatted=args.formatted
        )

    def add_arguments(self, parser):
        parser.add_argument(
            '--raw',
            help=(
                'Return the field value without applying '
                'plugin transformations'
            ),
            action='store_true',
            default=False
        )
        parser.add_argument(
            '--formatted',
            help=(
                'Format JSON output with indentation and sorted keys.'
            ),
            action='store_true',
            default=False
        )
        parser.add_argument(
            'field_name',
        )

    def get_field_value_by_dotpath(
        self, folder, field_name, raw=False, formatted=False
    ):
        fields = folder.get_fields()

        key_dotpath = None
        if '.' in field_name:
            field_name, key_dotpath = field_name.split('.', 1)

        if field_name not in fields:
            raise JirafsError("Field '%s' does not exist." % field_name)

        if raw:
            data = fields[field_name]
        else:
            data = fields.get_transformed(field_name)

        if key_dotpath:
            try:
                for component in key_dotpath.split('.'):
                    if not isinstance(data, dict):
                        raise JirafsError(
                            "Key '%s' (of dotpath '%s') is not an object "
                            "in field '%s'." % (
                                component,
                                key_dotpath,
                                field_name,
                            )
                        )
                    elif component not in data:
                        data = ''
                        break
                    else:
                        data = data[component]
            except (ValueError, TypeError):
                raise JirafsError(
                    "Field '%s' could not be parsed as JSON for retrieving "
                    "dotpath '%s'." % (
                        field_name,
                        key_dotpath,
                    )
                )

        return data

    def main(
        self, folder, field_name, raw=False, formatted=False
    ):
        data = self.get_field_value_by_dotpath(
            folder, field_name, raw, formatted
        )

        if isinstance(data, (list, dict)):
            kwargs = {}
            if formatted:
                kwargs = {
                    'indent': 4,
                    'sort_keys': True,
                }
            data = json.dumps(data, **kwargs)

        return data

    def cmd(self, *args, **kwargs):
        data = self.main(*args, **kwargs)

        return CommandResult(data, no_format=True)



from jirafs.plugin import CommandPlugin


class Command(CommandPlugin):
    """ Open a debug console """
    MIN_VERSION = '1.15'
    MAX_VERSION = '1.99.99'

    def main(self, folder, **kwargs):
        import ipdb
        return ipdb.set_trace()



import io
import json
import os
from unittest import TestCase

from mock import patch

from jira.resources import Issue


class BaseTestCase(TestCase):
    def get_asset_path(self, filename):
        return os.path.join(
            os.path.dirname(__file__),
            'assets',
            filename
        )

    def get_asset_contents(self, filename, mode='r'):
        path = self.get_asset_path(filename)

        with io.open(path, mode, encoding='utf-8') as r:
            return r.read()

    def rehydrate_issue(self, filename):
        stored = json.loads(self.get_asset_contents(filename))
        return Issue(
            stored['options'],
            None,
            stored['raw'],
        )

    def get_empty_status(self):
        return {
            'ready': {
                'files': [],
                'links': {},
                'fields': {},
                'new_comment': '',
            },
            'uncommitted': {
                'files': [],
                'links': {},
                'fields': {},
                'new_comment': '',
            },
            'local_uncommitted': {
                'files': [],
            },
            'up_to_date': True,
        }



import json

from jirafs.plugin import CommandPlugin, CommandResult


class Command(CommandPlugin):
    """ Get the status of the current ticketfolder """
    TRY_SUBFOLDERS = True
    MIN_VERSION = '1.15'
    MAX_VERSION = '1.99.99'

    def handle(self, args, folder, **kwargs):
        return self.cmd(folder, args.format)

    def add_arguments(self, parser):
        parser.add_argument(
            '--format',
            default='text',
            choices=['text', 'json']
        )

    def main(self, folder, output_format='text'):
        return folder.status()

    def cmd(self, folder, output_format='text'):
        status = self.main(folder)
        if output_format == 'json':
            return self.status_json(folder, status)
        return self.status_text(folder, status)

    def status_json(self, folder, status):
        result = CommandResult()
        result = result.add_line(json.dumps(status))
        return result

    def has_changes(self, section, *keys):
        if not keys:
            keys = ['files', 'fields', 'new_comment', 'links']
        for key in keys:
            if section[key]:
                return True

    def status_text(self, folder, folder_status):
        result = CommandResult()

        result = result.add_line(
            u"On ticket {ticket} ({url})",
            ticket=folder.ticket_number,
            url=folder.cached_issue.permalink(),
        )
        if not folder_status['up_to_date']:
            result = result.add_line(
                "{t.magenta}Warning: unmerged upstream changes exist; "
                "run `jirafs merge` to merge them into your local copy."
                "{t.normal}"
            )

        printed_changes = False
        ready = folder_status['ready']
        if self.has_changes(ready):
            printed_changes = True
            result = result.add_line('')
            result = result.add_line(
                "Ready for upload; use `jirafs push` to update JIRA."
            )
            result = self.format_field_changes(ready, 'green', result=result)

        staged = folder_status['uncommitted']
        if self.has_changes(staged):
            printed_changes = True
            result = result.add_line('')
            result = result.add_line(
                "Uncommitted changes; use `jirafs commit` to mark these "
                "for sending to JIRA during your next push, or "
                "`jirafs submit` to send these changes directly to JIRA."
            )
            result = self.format_field_changes(staged, 'red', result=result)

        local_uncommitted = folder_status['local_uncommitted']
        if self.has_changes(local_uncommitted, 'files'):
            printed_changes = True
            result = result.add_line('')
            result = result.add_line(
                "Uncommitted changes prevented from being sent to JIRA "
                "because they match at least one of the patterns in your "
                "`.jirafs_local` file; use `jirafs commit` to track these "
                "changes."
            )
            result = result.add_line(
                "Note: these files will {t.bold}not{t.normal} "
                "be uploaded to JIRA even after being committed."
            )
            result = self.format_field_changes(
                local_uncommitted,
                'cyan',
                no_upload=True,
                result=result,
            )

        if not printed_changes:
            result = result.add_line('No changes found')
        else:
            result = result.add_line("")
            result = result.add_line(
                'For more detail about these changes, run `jirafs diff`'
            )

        return result

    def format_field_changes(
        self, changes, color, no_upload=False, result=None
    ):
        if result is None:
            result = CommandResult()

        lines = []

        for filename in changes.get('files', []):
            result = result.add_line(
                u'\t{t.%s}{filename}{t.normal} {post_message}' % color,
                filename=filename,
                post_message=(
                    '(track in repository)'
                    if no_upload else '(upload attachment)'
                )
            )
        for link, data in changes.get('links', {}).get('remote', {}).items():
            orig = data[0]
            new = data[1]
            if new is not None:
                if new.get('description'):
                    description = new['description']
                else:
                    description = '(Untitled)'

                result = result.add_line(
                    u'\t{t.%s}{description}: '
                    u'{link}{t.normal} {post_message}' % (
                        color,
                    ),
                    description=description,
                    link=link,
                    post_message = (
                        ' (changed remote link)'
                        if orig else ' (new remote link)'
                    )
                )
            else:
                if orig.get('description'):
                    description = orig['description']
                else:
                    description = '(Untitled)'

                result = result.add_line(
                    u'\t{t.%s}{description}: '
                    u'{link}{t.normal} {post_message}' % (
                        color,
                    ),
                    description=description,
                    link=link,
                    post_message = '(removed remote link)'
                )
        for link, data in changes.get('links', {}).get('issue', {}).items():
            orig = data[0]
            new = data[1]
            if new is not None:
                if new.get('status'):
                    status = new['status']
                else:
                    status = '(Untitled)'
                result = result.add_line(
                    u'\t{t.%s}{status}: {link}{t.normal} {post_message}' % (
                        color
                    ),
                    status=status.title(),
                    link=link,
                    post_message=(
                        '(changed issue link)'
                        if orig else '(new issue link)'
                    )
                )
            else:
                if orig.get('status'):
                    status = orig['status']
                else:
                    status = '(Untitled)'

                result = result.add_line(
                    '\t{t.%s}{status}: {link}{t.normal} {post_message}' % (
                        color,
                    ),
                    status=status.title(),
                    link=link,
                    post_message='(removed issue link)'
                )
        for field, value_set in changes.get('fields', {}).items():
            result = result.add_line(
                '\t{t.%s}{field}{t.normal}' % color,
                field = field,
            )
        if changes.get('new_comment', ''):
            result = result.add_line(
                '\t{t.%s}[New Comment]{t.normal}' % color
            )

        return result



import io
from unittest import SkipTest

from mock import call, Mock, patch

from jirafs.utils import run_command_method_with_kwargs

from .base import BaseCommandTestCase


class TestPushCommand(BaseCommandTestCase):
    def setUp(self):
        super(TestPushCommand, self).setUp()

    def test_push_no_changes(self):
        with patch.object(self.ticketfolder.issue, 'update') as update:
            with patch('jirafs.commands.pull.Command.main') as pull:
                pull.return_value = True, True
                run_command_method_with_kwargs(
                    'push',
                    folder=self.ticketfolder,
                )

            self.assertEqual(0, len(update.call_args_list))

    def test_push_basic(self):
        status_result = {
            'ready': {
                'files': [],
                'fields': {
                    'somefield': ('one', 'two', 'two', ),
                    'otherfield': ('one', 'three', 'three', )
                },
                'links': {},
            }
        }

        with patch.object(self.ticketfolder, 'status') as status:
            status.return_value = status_result
            with patch.object(self.ticketfolder.issue, 'update') as update:
                with patch('jirafs.commands.pull.Command.main') as pull:
                    pull.return_value = True, True
                    run_command_method_with_kwargs(
                        'push',
                        folder=self.ticketfolder,
                    )

                self.assertEqual(1, len(update.call_args_list))
                self.assertEqual(
                    update.call_args,
                    call(
                        **{
                            field_name: value[1]
                            for field_name, value
                            in status_result['ready']['fields'].items()
                        }
                    )
                )


class TestPushCommandWithMacropatch(BaseCommandTestCase):
    def setUp(self):
        try:
            import jirafs_list_table  # noqa
        except ImportError:
            raise SkipTest(
                "Push command macropatch tests require the "
                "jira-list-table package to be installed."
            )
        super(TestPushCommandWithMacropatch, self).setUp()
        run_command_method_with_kwargs(
            'plugins',
            folder=self.ticketfolder,
            args=Mock(
                enable='list_table',
            )
        )

    def test_push_change_patched_content(self):
        # First, let's write out a patch
        description_one = u"""
            {list-table}
            * -
            ** Location
            ** Company
            * @coddingtonbear
            ** Portland, OR
            ** Urban Airship
            {list-table}
        """
        description_path = self.ticketfolder.get_path('description.jira')

        with io.open(description_path, 'w', encoding='utf-8') as out:
            out.write(description_one)

        run_command_method_with_kwargs(
            'commit',
            folder=self.ticketfolder,
            message='No me importa',
        )
        with patch.object(self.ticketfolder.issue, 'update'):
            with patch('jirafs.commands.pull.Command.pull') as pull:
                pull.return_value = True, True
                run_command_method_with_kwargs(
                    'push',
                    folder=self.ticketfolder,
                )

        description_two = u"""
            {list-table:horizontal}
            * -
            ** Location
            ** Company
            * @coddingtonbear
            ** Portland, OR
            ** Urban Airship
            * @ralphbean
            ** New York, NY
            ** RedHat
            {list-table}
        """

        with io.open(description_path, 'w', encoding='utf-8') as out:
            out.write(description_two)

        run_command_method_with_kwargs(
            'commit',
            folder=self.ticketfolder,
            message='No me importa',
        )
        with patch.object(self.ticketfolder.issue, 'update'):
            with patch('jirafs.commands.pull.Command.pull') as pull:
                pull.return_value = True, True
                run_command_method_with_kwargs(
                    'push',
                    folder=self.ticketfolder,
                )




from jirafs.plugin import CommandPlugin
from jirafs.utils import run_command_method_with_kwargs


class Command(CommandPlugin):
    """Commit current changes, push changes to JIRA, and pull changes"""
    MIN_VERSION = '1.15'
    MAX_VERSION = '1.99.99'

    def add_arguments(self, parser):
        parser.add_argument(
            '-m', '--message', dest='message', default='Untitled'
        )

    def handle(self, args, folder, **kwargs):
        return self.cmd(folder, args.message)

    def main(self, folder, message):
        commit_result = run_command_method_with_kwargs(
            'commit', folder=folder, message=message
        )
        push_result = run_command_method_with_kwargs(
            'push', folder=folder
        )

        return commit_result, push_result



import io
import json
import logging
import os
import re

import six

from jirafs import constants, utils
from jirafs.readers import GitRevisionReader, WorkingCopyReader


class JiraFieldManager(dict):
    FIELD_MATCHER = re.compile(
        '^%s$' % (
            constants.TICKET_FILE_FIELD_TEMPLATE.replace(
                '{field_name}', '([\w_]+)'
            )
        )
    )

    def __init__(self, data=None, names=None):
        if names is None:
            self._data, self._names = self.get_fields_from_string(data)
        else:
            self._data = data
            self._names = names
        super(JiraFieldManager, self).__init__(self._data)

    def __sub__(self, other):
        differing = {}
        for k, v in other.items():
            if (self.get(k) or v) and self.get(k) != v:
                tx = self.get(k)
                differing[k] = (v, tx, self.get(k), )

        return differing

    def get_human_name_for_field(self, field):
        try:
            return self._names[field]
        except KeyError:
            return field

    def items_transformed(self):
        for k, v in self.items():
            if k in self.get_requested_per_ticket_fields():
                result = self.folder.process_macros(v)
                yield k, result
            else:
                yield k, v

    def get_transformed(self, field_name, default=None):
        if field_name not in self.get_requested_per_ticket_fields():
            return self[field_name]

        try:
            return self.folder.process_macros(self[field_name])
        except KeyError:
            return default

    @classmethod
    def create(cls, folder, revision=None, path=None):
        if revision and path:
            raise TypeError(
                'You may specify a git revision or a local path; not both.'
            )

        if revision:
            return GitRevisionJiraFieldManager(folder, revision)
        else:
            return WorkingCopyJiraFieldManager(folder, path)

    def get_requested_per_ticket_fields(self):
        return constants.FILE_FIELDS

    def get_used_per_ticket_fields(self):
        raise NotImplementedError()

    def get_field_data_files(self):
        all_files = [
            constants.TICKET_DETAILS,
        ]

        used_per_ticket_fields = []
        try:
            for field in self.get_used_per_ticket_fields():
                all_files.append(
                    constants.TICKET_FILE_FIELD_TEMPLATE.format(
                        field_name=field,
                    )
                )
        except NotImplementedError:
            pass

        for field in self.get_requested_per_ticket_fields():
            all_files.append(
                constants.TICKET_FILE_FIELD_TEMPLATE.format(
                    field_name=field,
                )
            )

        return all_files

    def set_data_value(self, data, field_name, raw_value):
        raw_value = raw_value.strip()
        try:
            value = json.loads(raw_value)
        except (TypeError, ValueError):
            value = raw_value
        data[field_name] = value

    def get_fields_from_string(self, string):
        """ Gets field data from an incoming string.

        Parses through the string using the following RST-derived
        pattern::

            0 | * Field
            1 |     VALUE
            2 |     MORE VALUE

        """
        data = {}
        field_name = ''
        human_names = {}
        value = ''
        if not string:
            return data, human_names
        lines = string.split('\n')
        for idx, line in enumerate(lines):
            if line.startswith('*'):
                if value:  # If so, we just need to store previous loop data
                    self.set_data_value(data, field_name, value)
                    value = ''
                raw_field_name = re.match('^\* (.*\(.*?\)):$', line).group(1)
                if '(' in raw_field_name:
                    # This field name's real name doesn't match the field ID
                    match = re.match(
                        '(.*) \(([^)]+)\)',
                        raw_field_name
                    )
                    field_name = match.group(2)
                    human_names[field_name] = match.group(1)
                else:
                    field_name = raw_field_name.replace(' ', '_')
            elif field_name:
                value = value + '\n' + line.strip()
        if value:
            self.set_data_value(data, field_name, value)

        return data, human_names


class AutomaticJiraFieldManager(JiraFieldManager):
    def __init__(self):
        data, names = self.load()
        super(AutomaticJiraFieldManager, self).__init__(
            data, names=names
        )

    def load(self):
        fields, names = self.get_fields_from_string(
            self.get_file_contents(constants.TICKET_DETAILS)
        )

        used_fields = set(self.get_used_per_ticket_fields())
        requested_fields = set(self.get_requested_per_ticket_fields())

        file_fields = {}
        for field_name in used_fields | requested_fields:
            try:
                field_path = constants.TICKET_FILE_FIELD_TEMPLATE.format(
                    field_name=field_name
                )
                file_fields[field_name] = self.get_file_contents(field_path)
            except (IOError, OSError):
                pass

        fields.update(file_fields)
        return fields, names

    def get_file_contents(self, path):
        raise NotImplemented()


class WorkingCopyJiraFieldManager(
    WorkingCopyReader, AutomaticJiraFieldManager
):
    def get_used_per_ticket_fields(self):
        fields = []
        for filename in os.listdir(self.path):
            full_path = os.path.join(self.path, filename)
            matched = self.FIELD_MATCHER.match(filename)
            if matched and os.path.isfile(full_path):
                field_name = matched.group(1)
                if field_name not in constants.FILE_FIELD_BLACKLIST:
                    fields.append(field_name)

        return fields

    def write(self):
        folder_path = self.folder.get_path(constants.TICKET_DETAILS)

        used_fields = set(self.get_used_per_ticket_fields())
        requested_fields = set(self.get_requested_per_ticket_fields())

        with io.open(folder_path, 'w', encoding='utf-8') as out:
            for field in sorted(self.keys()):
                if field not in used_fields | requested_fields:
                    out.write(
                        u'* {human} ({field}):\n'.format(
                            human=self.get_human_name_for_field(field),
                            field=field,
                        )
                    )

                    field_string = self[field]
                    if field_string is None:
                        field_string = ''
                    elif not isinstance(field_string, six.string_types):
                        field_string = json.dumps(
                            field_string,
                            sort_keys=True,
                            indent=4,
                            ensure_ascii=False,
                        )

                    # Each line is preceded by 4 spaces of whitespace
                    padded_lines = [
                        u'    %s\n' % f for f in field_string.split('\n')
                    ]
                    for line in padded_lines:
                        out.write(line)
                else:
                    field_path = constants.TICKET_FILE_FIELD_TEMPLATE.format(
                        field_name=field
                    )
                    with io.open(field_path, 'w', encoding='utf-8') as fout:
                        fout.write(self[field])
                        fout.write(u'\n')


class GitRevisionJiraFieldManager(
    GitRevisionReader, AutomaticJiraFieldManager
):
    def get_used_per_ticket_fields(self):
        files = self.folder.run_git_command(
            'ls-tree',
            '--name-only',
            self.revision
        ).split()
        fields = []

        for filename in files:
            matched = self.FIELD_MATCHER.match(filename)
            if matched:
                field_name = matched.group(1)
                if field_name not in constants.FILE_FIELD_BLACKLIST:
                    fields.append(field_name)

        return fields



from jirafs.plugin import CommandPlugin
from jirafs.utils import run_command_method_with_kwargs


class Command(CommandPlugin):
    """ Fetch and merge remote changes """
    RUN_FOR_SUBTASKS = True
    TRY_SUBFOLDERS = True
    MIN_VERSION = '1.15'
    MAX_VERSION = '1.99.99'

    def main(self, folder, **kwargs):
        fetch_result = run_command_method_with_kwargs('fetch', folder=folder)
        merge_result = run_command_method_with_kwargs('merge', folder=folder)

        return fetch_result, merge_result
